/**
 * virtual-core
 *
 * Copyright (c) TanStack
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.md file in the root directory of this source tree.
 *
 * @license MIT
 */
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var observeRect_esm = require('../../../node_modules/@reach/observe-rect/dist/observe-rect.esm.js');
var utils = require('./utils.js');

//
const defaultKeyExtractor = index => index;
const defaultRangeExtractor = range => {
  const start = Math.max(range.startIndex - range.overscan, 0);
  const end = Math.min(range.endIndex + range.overscan, range.count - 1);
  const arr = [];

  for (let i = start; i <= end; i++) {
    arr.push(i);
  }

  return arr;
};
const observeElementRect = (instance, cb) => {
  const observer = observeRect_esm["default"](instance.scrollElement, rect => {
    cb(rect);
  });

  if (!instance.scrollElement) {
    return;
  }

  cb(instance.scrollElement.getBoundingClientRect());
  observer.observe();
  return () => {
    observer.unobserve();
  };
};
const observeWindowRect = (instance, cb) => {
  const onResize = () => {
    cb({
      width: instance.scrollElement.innerWidth,
      height: instance.scrollElement.innerHeight
    });
  };

  if (!instance.scrollElement) {
    return;
  }

  onResize();
  instance.scrollElement.addEventListener('resize', onResize, {
    capture: false,
    passive: true
  });
  return () => {
    instance.scrollElement.removeEventListener('resize', onResize);
  };
};
const observeElementOffset = (instance, cb) => {
  const onScroll = () => cb(instance.scrollElement[instance.options.horizontal ? 'scrollLeft' : 'scrollTop']);

  if (!instance.scrollElement) {
    return;
  }

  onScroll();
  instance.scrollElement.addEventListener('scroll', onScroll, {
    capture: false,
    passive: true
  });
  return () => {
    instance.scrollElement.removeEventListener('scroll', onScroll);
  };
};
const observeWindowOffset = (instance, cb) => {
  const onScroll = () => cb(instance.scrollElement[instance.options.horizontal ? 'scrollX' : 'scrollY']);

  if (!instance.scrollElement) {
    return;
  }

  onScroll();
  instance.scrollElement.addEventListener('scroll', onScroll, {
    capture: false,
    passive: true
  });
  return () => {
    instance.scrollElement.removeEventListener('scroll', onScroll);
  };
};
const measureElement = (element, instance) => {
  return element.getBoundingClientRect()[instance.options.horizontal ? 'width' : 'height'];
};
const windowScroll = (offset, canSmooth, instance) => {
  var _instance$scrollEleme;
  (_instance$scrollEleme = instance.scrollElement) == null ? void 0 : _instance$scrollEleme.scrollTo({
    [instance.options.horizontal ? 'left' : 'top']: offset,
    behavior: canSmooth ? 'smooth' : undefined
  });
};
const elementScroll = (offset, canSmooth, instance) => {
  var _instance$scrollEleme2;
  (_instance$scrollEleme2 = instance.scrollElement) == null ? void 0 : _instance$scrollEleme2.scrollTo({
    [instance.options.horizontal ? 'left' : 'top']: offset,
    behavior: canSmooth ? 'smooth' : undefined
  });
};
class Virtualizer {
  constructor(_opts) {
    var _this = this;

    this.unsubs = [];
    this.scrollElement = null;
    this.measurementsCache = [];
    this.itemMeasurementsCache = {};
    this.pendingMeasuredCacheIndexes = [];

    this.setOptions = opts => {
      Object.entries(opts).forEach(_ref => {
        let [key, value] = _ref;
        if (typeof value === 'undefined') delete opts[key];
      });
      this.options = {
        debug: false,
        initialOffset: 0,
        overscan: 1,
        paddingStart: 0,
        paddingEnd: 0,
        scrollPaddingStart: 0,
        scrollPaddingEnd: 0,
        horizontal: false,
        getItemKey: defaultKeyExtractor,
        rangeExtractor: defaultRangeExtractor,
        enableSmoothScroll: true,
        onChange: () => {},
        measureElement,
        initialRect: {
          width: 0,
          height: 0
        },
        ...opts
      };
    };

    this.notify = () => {
      var _this$options$onChang, _this$options;

      (_this$options$onChang = (_this$options = this.options).onChange) == null ? void 0 : _this$options$onChang.call(_this$options, this);
    };

    this.cleanup = () => {
      this.unsubs.filter(Boolean).forEach(d => d());
      this.unsubs = [];
    };

    this._didMount = () => {
      return () => {
        this.cleanup();
        this.scrollElement = null;
      };
    };

    this._willUpdate = () => {
      const scrollElement = this.options.getScrollElement();

      if (this.scrollElement !== scrollElement) {
        this.cleanup();
        this.scrollElement = scrollElement;
        this.unsubs.push(this.options.observeElementRect(this, rect => {
          this.scrollRect = rect;
          this.notify();
        }));
        this.unsubs.push(this.options.observeElementOffset(this, offset => {
          this.scrollOffset = offset;
          this.notify();
        }));
      }
    };

    this.getSize = () => {
      return this.scrollRect[this.options.horizontal ? 'width' : 'height'];
    };

    this.getMeasurements = utils.memo(() => [this.options.count, this.options.paddingStart, this.options.getItemKey, this.itemMeasurementsCache], (count, paddingStart, getItemKey, measurementsCache) => {
      const min = this.pendingMeasuredCacheIndexes.length > 0 ? Math.min(...this.pendingMeasuredCacheIndexes) : 0;
      this.pendingMeasuredCacheIndexes = [];
      const measurements = this.measurementsCache.slice(0, min);

      for (let i = min; i < count; i++) {
        const key = getItemKey(i);
        const measuredSize = measurementsCache[key];
        const start = measurements[i - 1] ? measurements[i - 1].end : paddingStart;
        const size = typeof measuredSize === 'number' ? measuredSize : this.options.estimateSize(i);
        const end = start + size;
        measurements[i] = {
          index: i,
          start,
          size,
          end,
          key
        };
      }

      this.measurementsCache = measurements;
      return measurements;
    }, {
      key: process.env.NODE_ENV === 'development' && 'getMeasurements',
      debug: () => this.options.debug
    });
    this.calculateRange = utils.memo(() => [this.getMeasurements(), this.getSize(), this.scrollOffset], (measurements, outerSize, scrollOffset) => {
      return calculateRange({
        measurements,
        outerSize,
        scrollOffset
      });
    }, {
      key: process.env.NODE_ENV === 'development' && 'calculateRange',
      debug: () => this.options.debug
    });
    this.getIndexes = utils.memo(() => [this.options.rangeExtractor, this.calculateRange(), this.options.overscan, this.options.count], (rangeExtractor, range, overscan, count) => {
      return rangeExtractor({ ...range,
        overscan,
        count: count
      });
    }, {
      key: process.env.NODE_ENV === 'development' && 'getIndexes'
    });
    this.getVirtualItems = utils.memo(() => [this.getIndexes(), this.getMeasurements(), this.options.measureElement], (indexes, measurements, measureElement) => {
      const virtualItems = [];

      for (let k = 0, len = indexes.length; k < len; k++) {
        const i = indexes[k];
        const measurement = measurements[i];
        const item = { ...measurement,
          measureElement: measurableItem => {
            if (measurableItem) {
              const measuredItemSize = measureElement(measurableItem, this);

              if (measuredItemSize !== item.size) {
                if (item.start < this.scrollOffset) {
                  if (process.env.NODE_ENV === 'development' && this.options.debug) console.info('correction', measuredItemSize - item.size);

                  if (!this.destinationOffset) {
                    this._scrollToOffset(this.scrollOffset + (measuredItemSize - item.size), false);
                  }
                }

                this.pendingMeasuredCacheIndexes.push(i);
                this.itemMeasurementsCache = { ...this.itemMeasurementsCache,
                  [item.key]: measuredItemSize
                };
                this.notify();
              }
            }
          }
        };
        virtualItems.push(item);
      }

      return virtualItems;
    }, {
      key: process.env.NODE_ENV === 'development' && 'getIndexes'
    });

    this.scrollToOffset = function (toOffset, _temp) {
      let {
        align
      } = _temp === void 0 ? {
        align: 'start'
      } : _temp;

      const attempt = () => {
        const offset = _this.scrollOffset;

        const size = _this.getSize();

        if (align === 'auto') {
          if (toOffset <= offset) {
            align = 'start';
          } else if (toOffset >= offset + size) {
            align = 'end';
          } else {
            align = 'start';
          }
        }

        if (align === 'start') {
          _this._scrollToOffset(toOffset, true);
        } else if (align === 'end') {
          _this._scrollToOffset(toOffset - size, true);
        } else if (align === 'center') {
          _this._scrollToOffset(toOffset - size / 2, true);
        }
      };

      attempt();
      requestAnimationFrame(() => {
        attempt();
      });
    };

    this.scrollToIndex = function (index, _temp2) {
      let {
        align,
        ...rest
      } = _temp2 === void 0 ? {
        align: 'auto'
      } : _temp2;

      const measurements = _this.getMeasurements();

      const offset = _this.scrollOffset;

      const size = _this.getSize();

      const {
        count
      } = _this.options;
      const measurement = measurements[Math.max(0, Math.min(index, count - 1))];

      if (!measurement) {
        return;
      }

      if (align === 'auto') {
        if (measurement.end >= offset + size - _this.options.scrollPaddingEnd) {
          align = 'end';
        } else if (measurement.start <= offset + _this.options.scrollPaddingStart) {
          align = 'start';
        } else {
          return;
        }
      }

      const toOffset = align === 'end' ? measurement.end + _this.options.scrollPaddingEnd : measurement.start - _this.options.scrollPaddingStart;

      _this.scrollToOffset(toOffset, {
        align,
        ...rest
      });
    };

    this.getTotalSize = () => {
      var _this$getMeasurements;

      return (((_this$getMeasurements = this.getMeasurements()[this.options.count - 1]) == null ? void 0 : _this$getMeasurements.end) || this.options.paddingStart) + this.options.paddingEnd;
    };

    this._scrollToOffset = (offset, canSmooth) => {
      clearTimeout(this.scrollCheckFrame);
      this.destinationOffset = offset;
      this.options.scrollToFn(offset, this.options.enableSmoothScroll && canSmooth, this);
      let scrollCheckFrame;

      const check = () => {
        let lastOffset = this.scrollOffset;
        this.scrollCheckFrame = scrollCheckFrame = setTimeout(() => {
          if (this.scrollCheckFrame !== scrollCheckFrame) {
            return;
          }

          if (this.scrollOffset === lastOffset) {
            this.destinationOffset = undefined;
            return;
          }

          lastOffset = this.scrollOffset;
          check();
        }, 100);
      };

      check();
    };

    this.measure = () => {
      this.itemMeasurementsCache = {};
      this.notify();
    };

    this.setOptions(_opts);
    this.scrollRect = this.options.initialRect;
    this.scrollOffset = this.options.initialOffset;
  }

}

const findNearestBinarySearch = (low, high, getCurrentValue, value) => {
  while (low <= high) {
    const middle = (low + high) / 2 | 0;
    const currentValue = getCurrentValue(middle);

    if (currentValue < value) {
      low = middle + 1;
    } else if (currentValue > value) {
      high = middle - 1;
    } else {
      return middle;
    }
  }

  if (low > 0) {
    return low - 1;
  } else {
    return 0;
  }
};

function calculateRange(_ref2) {
  let {
    measurements,
    outerSize,
    scrollOffset
  } = _ref2;
  const count = measurements.length - 1;

  const getOffset = index => measurements[index].start;

  const startIndex = findNearestBinarySearch(0, count, getOffset, scrollOffset);
  let endIndex = startIndex;

  while (endIndex < count && measurements[endIndex].end < scrollOffset + outerSize) {
    endIndex++;
  }

  return {
    startIndex,
    endIndex
  };
}

exports.memo = utils.memo;
exports.Virtualizer = Virtualizer;
exports.defaultKeyExtractor = defaultKeyExtractor;
exports.defaultRangeExtractor = defaultRangeExtractor;
exports.elementScroll = elementScroll;
exports.measureElement = measureElement;
exports.observeElementOffset = observeElementOffset;
exports.observeElementRect = observeElementRect;
exports.observeWindowOffset = observeWindowOffset;
exports.observeWindowRect = observeWindowRect;
exports.windowScroll = windowScroll;
//# sourceMappingURL=index.js.map
